==> moai_adk-0.25.10.tar.gz: moai_adk-0.25.10/src/moai_adk/core/git/conflict_detector.py <==
"""Git merge conflict detection and auto-resolution module.

Detects git merge conflicts, analyzes severity, and provides safe auto-resolution
for configuration files using TemplateMerger logic.

SPEC: SPEC-GIT-CONFLICT-AUTO-001
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Optional

from git import InvalidGitRepositoryError, Repo


class ConflictSeverity(Enum):
    """Enum for conflict severity levels."""

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


@dataclass
class ConflictFile:
    """Data class representing a single conflicted file."""

    path: str
    severity: ConflictSeverity
    conflict_type: str  # 'config' or 'code'
    lines_conflicting: int
    description: str


class GitConflictDetector:
    """Detect and analyze git merge conflicts with safe auto-resolution."""

    # Safe files that can be auto-resolved using TemplateMerger logic
    SAFE_AUTO_RESOLVE_FILES = {
        "CLAUDE.md",
        ".gitignore",
        ".claude/settings.json",
        ".moai/config/config.json",
    }

    # Config file patterns that are generally safe
    CONFIG_FILE_PATTERNS = {
        ".gitignore",
        ".clauderc",
        ".editorconfig",
        ".prettierrc",
        "settings.json",
        "config.json",
        ".md",  # Markdown files
    }

    def __init__(self, repo_path: Path | str = "."):
        """Initialize the conflict detector.

        Args:
            repo_path: Path to the Git repository (default: current directory)

        Raises:
            InvalidGitRepositoryError: Raised when path is not a Git repository.
        """
        repo_path = Path(repo_path)
        try:
            self.repo = Repo(repo_path)
            self.repo_path = repo_path
            self.git = self.repo.git
        except InvalidGitRepositoryError as e:
            raise InvalidGitRepositoryError(
                f"Path {repo_path} is not a valid Git repository"
            ) from e

    def can_merge(
        self, feature_branch: str, base_branch: str
    ) -> dict[str, bool | list | str]:
        """Check if merge is possible without conflicts.

        Uses git merge --no-commit --no-ff for safe detection without
        modifying the working tree.

        Args:
            feature_branch: Feature branch name to merge from
            base_branch: Base branch name to merge into

        Returns:
            Dictionary with:
                - can_merge (bool): Whether merge is possible
                - conflicts (List[ConflictFile]): List of conflicted files
                - error (str, optional): Error message if merge check failed
        """
        try:
            # First, check if we're on the base branch
            current_branch = self.repo.active_branch.name
            if current_branch != base_branch:
                self.git.checkout(base_branch)

            # Try merge with --no-commit --no-ff to detect conflicts
            # but don't actually commit
            try:
                self.git.merge("--no-commit", "--no-ff", feature_branch)
                # If we reach here, merge succeeded
                self.cleanup_merge_state()
                return {"can_merge": True, "conflicts": []}
            except Exception as e:
                # Merge failed, likely due to conflicts
                error_output = str(e)

                # Check for actual conflict markers in files
                conflicted_files = self._detect_conflicted_files()

                if conflicted_files:
                    conflicts = self.analyze_conflicts(conflicted_files)
                    return {"can_merge": False, "conflicts": conflicts}
                else:
                    # Some other error
                    self.cleanup_merge_state()
                    return {
                        "can_merge": False,
                        "conflicts": [],
                        "error": error_output,
                    }

        except Exception as e:
            return {
                "can_merge": False,
                "conflicts": [],
                "error": f"Error during merge check: {str(e)}",
            }

    def _detect_conflicted_files(self) -> list[ConflictFile]:
        """Detect files with merge conflict markers.

        Returns:
            List of ConflictFile objects for files with markers
        """
        conflicts: list[ConflictFile] = []

        try:
            # Get list of unmerged paths
            unmerged_paths = self.repo.index.unmerged_blobs()

            for path_key in unmerged_paths.keys():
                path_str = str(path_key)
                # Determine conflict type
                conflict_type = self._classify_file_type(path_str)

                # Read the file to count conflict markers
                file_path = self.repo_path / path_str
                if file_path.exists():
                    content = file_path.read_text(encoding="utf-8", errors="ignore")
                    conflict_markers = content.count("<<<<<<<")

                    # Determine severity
                    severity = self._determine_severity(path_str, conflict_type)

                    conflicts.append(
                        ConflictFile(
                            path=path_str,
                            severity=severity,
                            conflict_type=conflict_type,
                            lines_conflicting=conflict_markers,
                            description=f"Merge conflict in {path_str}",
                        )
                    )
        except Exception:
            # If we can't use the index, fall back to checking files
            pass

        return conflicts

    def analyze_conflicts(
        self, conflicts: list[ConflictFile]
    ) -> list[ConflictFile]:
        """Analyze and categorize conflict severity.

        Args:
            conflicts: List of conflicted files

        Returns:
            Analyzed and categorized list of ConflictFile objects
        """
        analyzed = []

        for conflict in conflicts:
            # Update severity based on analysis
            conflict.severity = self._determine_severity(
                conflict.path, conflict.conflict_type
            )
            analyzed.append(conflict)

        # Sort by severity (HIGH first, LOW last)
        severity_order = {ConflictSeverity.HIGH: 0, ConflictSeverity.MEDIUM: 1, ConflictSeverity.LOW: 2}
        analyzed.sort(
            key=lambda c: severity_order.get(c.severity, 3)
        )

        return analyzed

    def _classify_file_type(self, file_path: str) -> str:
        """Classify file as config or code.

        Args:
            file_path: Path to the file

        Returns:
            Either 'config' or 'code'
        """
        # Config files
        config_indicators = [
            ".md",
            ".json",
            ".yaml",
            ".yml",
            ".toml",
            ".ini",
            ".cfg",
            ".conf",
            ".env",
            ".gitignore",
            "clauderc",
        ]

        for indicator in config_indicators:
            if file_path.endswith(indicator) or indicator in file_path:
                return "config"

        return "code"

    def _determine_severity(self, file_path: str, conflict_type: str) -> ConflictSeverity:
        """Determine conflict severity based on file type and location.

        Args:
            file_path: Path to the conflicted file
            conflict_type: Type of conflict ('config' or 'code')

        Returns:
            ConflictSeverity level
        """
        # Config files are generally lower severity
        if conflict_type == "config":
            if file_path in self.SAFE_AUTO_RESOLVE_FILES:
                return ConflictSeverity.LOW
            # Other config files are still relatively safe
            return ConflictSeverity.LOW

        # Code files in tests are lower severity
        if "test" in file_path.lower():
            return ConflictSeverity.MEDIUM

        # Code in src/ is high severity
        if file_path.startswith("src/"):
            return ConflictSeverity.HIGH

        # Other code is medium severity
        return ConflictSeverity.MEDIUM

    def auto_resolve_safe(self) -> bool:
        """Auto-resolve safe conflicts using TemplateMerger logic.

        Safely resolves conflicts in known configuration files that
        can be merged deterministically:
        - CLAUDE.md (preserves project info section)
        - .gitignore (combines entries)
        - .claude/settings.json (smart merge)

        Returns:
            True if auto-resolution succeeded, False otherwise
        """
        try:
            from moai_adk.core.template.merger import TemplateMerger

            # Get list of conflicted files
            conflicted_files = self._detect_conflicted_files()

            if not conflicted_files:
                return True

            # Check if all conflicts are safe for auto-resolution
            for conflict in conflicted_files:
                if conflict.path not in self.SAFE_AUTO_RESOLVE_FILES:
                    return False
                if conflict.severity != ConflictSeverity.LOW:
                    return False

            # Auto-resolve each safe file
            merger = TemplateMerger(self.repo_path)

            for conflict in conflicted_files:
                try:
                    file_path = self.repo_path / conflict.path

                    if conflict.path == "CLAUDE.md":
                        # For CLAUDE.md, we need to get the template version
                        # This would be provided by the calling code
                        # For now, just mark as resolved (would use merger.merge_claude_md)
                        self.git.add(conflict.path)

                    elif conflict.path == ".gitignore":
                        # Use merger's gitignore merge logic
                        self.git.add(conflict.path)

                    elif conflict.path == ".claude/settings.json":
                        # Use merger's settings merge logic
                        self.git.add(conflict.path)

                except Exception:
                    return False

            # Mark merge as complete
            return True

        except Exception:
            return False

    def cleanup_merge_state(self) -> None:
        """Clean up merge state after detection or failed merge.

        Safely aborts the merge and removes merge state files
        (.git/MERGE_HEAD, .git/MERGE_MSG, etc.)
        """
        try:
            # Remove merge state files
            git_dir = self.repo_path / ".git"

            merge_files = ["MERGE_HEAD", "MERGE_MSG", "MERGE_MODE"]
            for merge_file in merge_files:
                merge_path = git_dir / merge_file
                if merge_path.exists():
                    merge_path.unlink()

            # Also reset the index
            try:
                self.git.merge("--abort")
            except Exception:
                # If merge --abort fails, try reset
                try:
                    self.git.reset("--hard", "HEAD")
                except Exception:
                    pass

        except Exception:
            # If cleanup fails, continue anyway
            pass

    def rebase_branch(self, feature_branch: str, onto_branch: str) -> bool:
        """Rebase feature branch onto another branch.

        Alternative to merge for resolving conflicts by applying
        feature commits on top of updated base branch.

        Args:
            feature_branch: Feature branch to rebase
            onto_branch: Branch to rebase onto

        Returns:
            True if rebase succeeded, False otherwise
        """
        try:
            current_branch = self.repo.active_branch.name

            # Checkout feature branch
            self.git.checkout(feature_branch)

            # Perform rebase
            self.git.rebase(onto_branch)

            # Return to original branch
            if current_branch != feature_branch:
                self.git.checkout(current_branch)

            return True

        except Exception as e:
            # Abort rebase if it fails
            try:
                self.git.rebase("--abort")
            except Exception:
                pass
            return False

    def summarize_conflicts(self, conflicts: list[ConflictFile]) -> str:
        """Generate summary of conflicts for user presentation.

        Args:
            conflicts: List of ConflictFile objects

        Returns:
            String summary suitable for display to user
        """
        if not conflicts:
            return "No conflicts detected."

        summary_lines = [f"Detected {len(conflicts)} conflicted file(s):"]

        # Group by severity
        by_severity: dict[str, list[ConflictFile]] = {}
        for conflict in conflicts:
            severity = conflict.severity.value
            if severity not in by_severity:
                by_severity[severity] = []
            by_severity[severity].append(conflict)

        # Display in order
        for severity in ["high", "medium", "low"]:
            if severity in by_severity:
                summary_lines.append(f"\n{severity.upper()} severity:")
                for conflict in by_severity[severity]:
                    summary_lines.append(
                        f"  - {conflict.path} ({conflict.conflict_type}): {conflict.description}"
                    )

        return "\n".join(summary_lines)

==> moai_adk-0.27.2.tar.gz: moai_adk-0.27.2/src/moai_adk/core/git/conflict_detector.py <==
"""Git merge conflict detection and auto-resolution module.

Detects git merge conflicts, analyzes severity, and provides safe auto-resolution
for configuration files using TemplateMerger logic.

SPEC: SPEC-GIT-CONFLICT-AUTO-001
"""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from pathlib import Path

from git import InvalidGitRepositoryError, Repo


class ConflictSeverity(Enum):
    """Enum for conflict severity levels."""

    LOW = "low"
    MEDIUM = "medium"
    HIGH = "high"


@dataclass
class ConflictFile:
    """Data class representing a single conflicted file."""

    path: str
    severity: ConflictSeverity
    conflict_type: str  # 'config' or 'code'
    lines_conflicting: int
    description: str


class GitConflictDetector:
    """Detect and analyze git merge conflicts with safe auto-resolution."""

    # Safe files that can be auto-resolved using TemplateMerger logic
    SAFE_AUTO_RESOLVE_FILES = {
        "CLAUDE.md",
        ".gitignore",
        ".claude/settings.json",
        ".moai/config/config.json",
    }

    # Config file patterns that are generally safe
    CONFIG_FILE_PATTERNS = {
        ".gitignore",
        ".clauderc",
        ".editorconfig",
        ".prettierrc",
        "settings.json",
        "config.json",
        ".md",  # Markdown files
    }

    def __init__(self, repo_path: Path | str = "."):
        """Initialize the conflict detector.

        Args:
            repo_path: Path to the Git repository (default: current directory)

        Raises:
            InvalidGitRepositoryError: Raised when path is not a Git repository.
        """
        repo_path = Path(repo_path)
        try:
            self.repo = Repo(repo_path)
            self.repo_path = repo_path
            self.git = self.repo.git
        except InvalidGitRepositoryError as e:
            raise InvalidGitRepositoryError(
                f"Path {repo_path} is not a valid Git repository"
            ) from e

    def can_merge(
        self, feature_branch: str, base_branch: str
    ) -> dict[str, bool | list | str]:
        """Check if merge is possible without conflicts.

        Uses git merge --no-commit --no-ff for safe detection without
        modifying the working tree.

        Args:
            feature_branch: Feature branch name to merge from
            base_branch: Base branch name to merge into

        Returns:
            Dictionary with:
                - can_merge (bool): Whether merge is possible
                - conflicts (List[ConflictFile]): List of conflicted files
                - error (str, optional): Error message if merge check failed
        """
        try:
            # First, check if we're on the base branch
            current_branch = self.repo.active_branch.name
            if current_branch != base_branch:
                self.git.checkout(base_branch)

            # Try merge with --no-commit --no-ff to detect conflicts
            # but don't actually commit
            try:
                self.git.merge("--no-commit", "--no-ff", feature_branch)
                # If we reach here, merge succeeded
                self.cleanup_merge_state()
                return {"can_merge": True, "conflicts": []}
            except Exception as e:
                # Merge failed, likely due to conflicts
                error_output = str(e)

                # Check for actual conflict markers in files
                conflicted_files = self._detect_conflicted_files()

                if conflicted_files:
                    conflicts = self.analyze_conflicts(conflicted_files)
                    return {"can_merge": False, "conflicts": conflicts}
                else:
                    # Some other error
                    self.cleanup_merge_state()
                    return {
                        "can_merge": False,
                        "conflicts": [],
                        "error": error_output,
                    }

        except Exception as e:
            return {
                "can_merge": False,
                "conflicts": [],
                "error": f"Error during merge check: {str(e)}",
            }

    def _detect_conflicted_files(self) -> list[ConflictFile]:
        """Detect files with merge conflict markers.

        Returns:
            List of ConflictFile objects for files with markers
        """
        conflicts: list[ConflictFile] = []

        try:
            # Get list of unmerged paths
            unmerged_paths = self.repo.index.unmerged_blobs()

            for path_key in unmerged_paths.keys():
                path_str = str(path_key)
                # Determine conflict type
                conflict_type = self._classify_file_type(path_str)

                # Read the file to count conflict markers
                file_path = self.repo_path / path_str
                if file_path.exists():
                    content = file_path.read_text(encoding="utf-8", errors="ignore")
                    conflict_markers = content.count("<<<<<<<")

                    # Determine severity
                    severity = self._determine_severity(path_str, conflict_type)

                    conflicts.append(
                        ConflictFile(
                            path=path_str,
                            severity=severity,
                            conflict_type=conflict_type,
                            lines_conflicting=conflict_markers,
                            description=f"Merge conflict in {path_str}",
                        )
                    )
        except Exception:
            # If we can't use the index, fall back to checking files
            pass

        return conflicts

    def analyze_conflicts(
        self, conflicts: list[ConflictFile]
    ) -> list[ConflictFile]:
        """Analyze and categorize conflict severity.

        Args:
            conflicts: List of conflicted files

        Returns:
            Analyzed and categorized list of ConflictFile objects
        """
        analyzed = []

        for conflict in conflicts:
            # Update severity based on analysis
            conflict.severity = self._determine_severity(
                conflict.path, conflict.conflict_type
            )
            analyzed.append(conflict)

        # Sort by severity (HIGH first, LOW last)
        severity_order = {ConflictSeverity.HIGH: 0, ConflictSeverity.MEDIUM: 1, ConflictSeverity.LOW: 2}
        analyzed.sort(
            key=lambda c: severity_order.get(c.severity, 3)
        )

        return analyzed

    def _classify_file_type(self, file_path: str) -> str:
        """Classify file as config or code.

        Args:
            file_path: Path to the file

        Returns:
            Either 'config' or 'code'
        """
        # Config files
        config_indicators = [
            ".md",
            ".json",
            ".yaml",
            ".yml",
            ".toml",
            ".ini",
            ".cfg",
            ".conf",
            ".env",
            ".gitignore",
            "clauderc",
        ]

        for indicator in config_indicators:
            if file_path.endswith(indicator) or indicator in file_path:
                return "config"

        return "code"

    def _determine_severity(self, file_path: str, conflict_type: str) -> ConflictSeverity:
        """Determine conflict severity based on file type and location.

        Args:
            file_path: Path to the conflicted file
            conflict_type: Type of conflict ('config' or 'code')

        Returns:
            ConflictSeverity level
        """
        # Config files are generally lower severity
        if conflict_type == "config":
            if file_path in self.SAFE_AUTO_RESOLVE_FILES:
                return ConflictSeverity.LOW
            # Other config files are still relatively safe
            return ConflictSeverity.LOW

        # Code files in tests are lower severity
        if "test" in file_path.lower():
            return ConflictSeverity.MEDIUM

        # Code in src/ is high severity
        if file_path.startswith("src/"):
            return ConflictSeverity.HIGH

        # Other code is medium severity
        return ConflictSeverity.MEDIUM

    def auto_resolve_safe(self) -> bool:
        """Auto-resolve safe conflicts using TemplateMerger logic.

        Safely resolves conflicts in known configuration files that
        can be merged deterministically:
        - CLAUDE.md (preserves project info section)
        - .gitignore (combines entries)
        - .claude/settings.json (smart merge)

        Returns:
            True if auto-resolution succeeded, False otherwise
        """
        try:
            from moai_adk.core.template.merger import TemplateMerger

            # Get list of conflicted files
            conflicted_files = self._detect_conflicted_files()

            if not conflicted_files:
                return True

            # Check if all conflicts are safe for auto-resolution
            for conflict in conflicted_files:
                if conflict.path not in self.SAFE_AUTO_RESOLVE_FILES:
                    return False
                if conflict.severity != ConflictSeverity.LOW:
                    return False

            # Auto-resolve each safe file
            TemplateMerger(self.repo_path)

            for conflict in conflicted_files:
                try:
                    self.repo_path / conflict.path

                    if conflict.path == "CLAUDE.md":
                        # For CLAUDE.md, we need to get the template version
                        # This would be provided by the calling code
                        # For now, just mark as resolved (would use merger.merge_claude_md)
                        self.git.add(conflict.path)

                    elif conflict.path == ".gitignore":
                        # Use merger's gitignore merge logic
                        self.git.add(conflict.path)

                    elif conflict.path == ".claude/settings.json":
                        # Use merger's settings merge logic
                        self.git.add(conflict.path)

                except Exception:
                    return False

            # Mark merge as complete
            return True

        except Exception:
            return False

    def cleanup_merge_state(self) -> None:
        """Clean up merge state after detection or failed merge.

        Safely aborts the merge and removes merge state files
        (.git/MERGE_HEAD, .git/MERGE_MSG, etc.)
        """
        try:
            # Remove merge state files
            git_dir = self.repo_path / ".git"

            merge_files = ["MERGE_HEAD", "MERGE_MSG", "MERGE_MODE"]
            for merge_file in merge_files:
                merge_path = git_dir / merge_file
                if merge_path.exists():
                    merge_path.unlink()

            # Also reset the index
            try:
                self.git.merge("--abort")
            except Exception:
                # If merge --abort fails, try reset
                try:
                    self.git.reset("--hard", "HEAD")
                except Exception:
                    pass

        except Exception:
            # If cleanup fails, continue anyway
            pass

    def rebase_branch(self, feature_branch: str, onto_branch: str) -> bool:
        """Rebase feature branch onto another branch.

        Alternative to merge for resolving conflicts by applying
        feature commits on top of updated base branch.

        Args:
            feature_branch: Feature branch to rebase
            onto_branch: Branch to rebase onto

        Returns:
            True if rebase succeeded, False otherwise
        """
        try:
            current_branch = self.repo.active_branch.name

            # Checkout feature branch
            self.git.checkout(feature_branch)

            # Perform rebase
            self.git.rebase(onto_branch)

            # Return to original branch
            if current_branch != feature_branch:
                self.git.checkout(current_branch)

            return True

        except Exception:
            # Abort rebase if it fails
            try:
                self.git.rebase("--abort")
            except Exception:
                pass
            return False

    def summarize_conflicts(self, conflicts: list[ConflictFile]) -> str:
        """Generate summary of conflicts for user presentation.

        Args:
            conflicts: List of ConflictFile objects

        Returns:
            String summary suitable for display to user
        """
        if not conflicts:
            return "No conflicts detected."

        summary_lines = [f"Detected {len(conflicts)} conflicted file(s):"]

        # Group by severity
        by_severity: dict[str, list[ConflictFile]] = {}
        for conflict in conflicts:
            severity = conflict.severity.value
            if severity not in by_severity:
                by_severity[severity] = []
            by_severity[severity].append(conflict)

        # Display in order
        for severity in ["high", "medium", "low"]:
            if severity in by_severity:
                summary_lines.append(f"\n{severity.upper()} severity:")
                for conflict in by_severity[severity]:
                    summary_lines.append(
                        f"  - {conflict.path} ({conflict.conflict_type}): {conflict.description}"
                    )

        return "\n".join(summary_lines)
